@@ -43,11 +43,15 @@ table_mode:
     * direct - non-partitioned table
     * ignore - all events are ignored
 
+part_func:
+    database function to use for creating partition table.
+    default is {londiste|public}.create_partition
+
 part_mode:
-    * batch_time - paritioned by batch creation time (default)
+    * batch_time - partitioned by batch creation time (default)
     * event_time - partitioned by event creation time
-    * date_field - paritioned by date_field value. part_field required
-    * current_time - paritioned by current time
+    * date_field - partitioned by date_field value. part_field required
+    * current_time - partitioned by current time
 
 part_field:
     date_field to use for partition. Required when part_mode=date_field
@@ -72,7 +76,7 @@ part_name:
 
 part_template:
     custom sql template for creating partition table. if omitted then partition
-    function is used. when function is missing then parent table is cloned.
+    function is used.
     template variables:
     * dest - destination table name. result on part_name evaluation
     * part - same as dest
@@ -86,7 +90,7 @@ part_template:
 row_mode:
     how rows are applied to target table
     * plain - each event creates SQL statement to run (default)
-    * keep_latest - change updates to DELETE + INSERT, ignore deletes
+    * keep_latest - change updates to DELETE + INSERT
     * keep_all - change updates to inserts, ignore deletes
 
 event_types:
@@ -131,6 +135,17 @@ post_part:
     sql statement(s) to execute after creating partition table. Usable
     variables are the same as in part_template
 
+retention_period:
+    how long to keep partitions around. examples: '3 months', '1 year'
+
+ignore_old_events:
+    * 0 - handle all events in the same way (default)
+    * 1 - ignore events coming for obsolete partitions
+
+ignore_truncate:
+    * 0 - process truncate event (default)
+    * 1 - ignore truncate event
+
 encoding:
     name of destination encoding. handler replaces all invalid encoding symbols
     and logs them as warnings
@@ -146,17 +161,20 @@ creating or coping initial data to destination table.  --expect-sync and
 --skip-truncate should be used and --create switch is to be avoided.
 """
 
-import sys
-import datetime
 import codecs
+import datetime
 import re
+import sys
+from functools import partial
+
 import skytools
-from londiste.handler import BaseHandler, EncodingValidator
 from skytools import quote_ident, quote_fqident, UsageError
 from skytools.dbstruct import *
 from skytools.utf8 import safe_utf8_decode
-from functools import partial
+
+from londiste.handler import EncodingValidator
 from londiste.handlers import handler_args, update
+from londiste.handlers.shard import ShardHandler
 
 
 __all__ = ['Dispatcher']
@@ -186,6 +204,8 @@ PART_FUNC_OLD = 'public.create_partition'
 PART_FUNC_NEW = 'londiste.create_partition'
 PART_FUNC_ARGS = ['parent', 'part', 'pkeys', 'part_field', 'part_time', 'period']
 
+RETENTION_FUNC = "londiste.drop_obsolete_partitions"
+
 
 
 #------------------------------------------------------------------------------
@@ -221,7 +241,7 @@ class DirectLoader(BaseLoader):
                   'D': skytools.mk_delete_sql}
         if self.data:
             curs.execute("\n".join(mk_sql[op](row, self.table, self.pkeys)
-                for op, row in self.data))
+                                   for op, row in self.data))
 
 
 class BaseBulkCollectingLoader(BaseLoader):
@@ -307,10 +327,9 @@ class BaseBulkTempLoader(BaseBulkCollectingLoader):
 
     def logexec(self, curs, sql):
         """Logs and executes sql statement"""
-        self.log.debug('exec: %s' % sql)
+        self.log.debug('exec: %s', sql)
         curs.execute(sql)
-        self.log.debug('msg: %s, rows: %s' % (
-            curs.statusmessage, curs.rowcount))
+        self.log.debug('msg: %s, rows: %s', curs.statusmessage, curs.rowcount)
 
     # create sql parts
 
@@ -394,15 +413,15 @@ class BulkLoader(BaseBulkTempLoader):
         cnt = len(data)
         if (cnt == 0):
             return
-        self.log.debug("bulk: Deleting %d rows from %s" % (cnt, self.table))
+        self.log.debug("bulk: Deleting %d rows from %s", cnt, self.table)
         # copy rows to temp
         self.bulk_insert(curs, data)
         # delete rows using temp
         self.delete(curs)
         # check if right amount of rows deleted (only in direct mode)
         if self.conf.table_mode == 'direct' and cnt != curs.rowcount:
-            self.log.warning("%s: Delete mismatch: expected=%s deleted=%d"
-                    % (self.table, cnt, curs.rowcount))
+            self.log.warning("%s: Delete mismatch: expected=%s deleted=%d",
+                    self.table, cnt, curs.rowcount)
 
     def process_update(self, curs, op_map):
         """Process update list"""
@@ -415,7 +434,7 @@ class BulkLoader(BaseBulkTempLoader):
         cnt = len(data)
         if (cnt == 0):
             return
-        self.log.debug("bulk: Updating %d rows in %s" % (cnt, self.table))
+        self.log.debug("bulk: Updating %d rows in %s", cnt, self.table)
         # copy rows to temp
         self.bulk_insert(curs, data)
         if self.method == METH_CORRECT:
@@ -423,15 +442,15 @@ class BulkLoader(BaseBulkTempLoader):
             self.update(curs)
             # check count (only in direct mode)
             if self.conf.table_mode == 'direct' and cnt != curs.rowcount:
-                self.log.warning("%s: Update mismatch: expected=%s updated=%d"
-                        % (self.table, cnt, curs.rowcount))
+                self.log.warning("%s: Update mismatch: expected=%s updated=%d",
+                        self.table, cnt, curs.rowcount)
         else:
             # delete from main table using temp
             self.delete(curs)
             # check count (only in direct mode)
             if self.conf.table_mode == 'direct' and real_cnt != curs.rowcount:
-                self.log.warning("%s: Update mismatch: expected=%s deleted=%d"
-                        % (self.table, real_cnt, curs.rowcount))
+                self.log.warning("%s: Update mismatch: expected=%s deleted=%d",
+                        self.table, real_cnt, curs.rowcount)
             # insert into main table
             if AVOID_BIZGRES_BUG:
                 # copy again, into main table
@@ -448,19 +467,19 @@ class BulkLoader(BaseBulkTempLoader):
         # merged method loads inserts together with updates
         if (cnt == 0) or (self.method == METH_MERGED):
             return
-        self.log.debug("bulk: Inserting %d rows into %s" % (cnt, self.table))
+        self.log.debug("bulk: Inserting %d rows into %s", cnt, self.table)
         # copy into target table (no temp used)
         self.bulk_insert(curs, data, table = self.qtable)
 
     def bulk_flush(self, curs, op_map):
-        self.log.debug("bulk_flush: %s  (I/U/D = %d/%d/%d)" % (
-            self.table, len(op_map['I']), len(op_map['U']), len(op_map['D'])))
+        self.log.debug("bulk_flush: %s  (I/U/D = %d/%d/%d)", self.table,
+                len(op_map['I']), len(op_map['U']), len(op_map['D']))
 
         # fetch distribution fields
         if self.dist_fields is None:
             self.dist_fields = self.find_dist_fields(curs)
-            self.log.debug("Key fields: %s  Dist fields: %s" % (
-                           ",".join(self.pkeys), ",".join(self.dist_fields)))
+            self.log.debug("Key fields: %s  Dist fields: %s",
+                           ",".join(self.pkeys), ",".join(self.dist_fields))
             # add them to key
             for key in self.dist_fields:
                 if key not in self.keys:
@@ -496,7 +515,7 @@ class BulkLoader(BaseBulkTempLoader):
         """
         if USE_LONGLIVED_TEMP_TABLES or USE_REAL_TABLE:
             if self.temp_present:
-                self.log.debug("bulk: Using existing temp table %s" % self.temp)
+                self.log.debug("bulk: Using existing temp table %s", self.temp)
                 return False
         self.create(curs)
         self.temp_present = True
@@ -516,7 +535,7 @@ class BulkLoader(BaseBulkTempLoader):
             # truncate when re-using existing table
             if not self.create_temp(curs):
                 self.truncate(curs)
-        self.log.debug("bulk: COPY %d rows into %s" % (len(data), table))
+        self.log.debug("bulk: COPY %d rows into %s", len(data), table)
         skytools.magic_insert(curs, table, data, self.fields,
                               quoted_table = True)
         if _use_temp and self.run_analyze:
@@ -588,7 +607,7 @@ class KeepLatestRowHandler(RowHandler):
     def process(self, table, op, row):
         """Keep latest row version.
 
-        Updates are changed to delete + insert, deletes are ignored.
+        Updates are changed to delete + insert
         Makes sense only for partitioned tables.
         """
         if op == 'U':
@@ -596,6 +615,8 @@ class KeepLatestRowHandler(RowHandler):
             RowHandler.process(self, table, 'I', row)
         elif op == 'I':
             RowHandler.process(self, table, 'I', row)
+        elif op == 'D':
+            RowHandler.process(self, table, 'D', row)
 
 
 ROW_HANDLERS = {'plain': RowHandler,
@@ -608,7 +629,7 @@ ROW_HANDLERS = {'plain': RowHandler,
 #------------------------------------------------------------------------------
 
 
-class Dispatcher(BaseHandler):
+class Dispatcher (ShardHandler):
     """Partitioned loader.
     Splits events into partitions, if requested.
     Then applies them without further processing.
@@ -620,23 +641,17 @@ class Dispatcher(BaseHandler):
         # compat for dest-table
         dest_table = args.get('table', dest_table)
 
-        BaseHandler.__init__(self, table_name, args, dest_table)
+        ShardHandler.__init__(self, table_name, args, dest_table)
 
         # show args
-        self.log.debug("dispatch.init: table_name=%r, args=%r" % \
-                       (table_name, args))
+        self.log.debug("dispatch.init: table_name=%r, args=%r", table_name, args)
+        self.ignored_tables = set()
         self.batch_info = None
         self.dst_curs = None
         self.pkeys = None
         # config
-        self.conf = self.get_config()
         hdlr_cls = ROW_HANDLERS[self.conf.row_mode]
         self.row_handler = hdlr_cls(self.log)
-        if self.conf.encoding:
-            self.encoding_validator = EncodingValidator(self.log,
-                                                        self.conf.encoding)
-        else:
-            self.encoding_validator = None
 
     def _parse_args_from_doc (self):
         doc = __doc__
@@ -663,7 +678,7 @@ class Dispatcher(BaseHandler):
 
     def get_config(self):
         """Processes args dict"""
-        conf = skytools.dbdict()
+        conf = ShardHandler.get_config(self)
         # set table mode
         conf.table_mode = self.get_arg('table_mode', TABLE_MODES)
         conf.analyze = self.get_arg('analyze', [0, 1])
@@ -678,6 +693,8 @@ class Dispatcher(BaseHandler):
             conf.pre_part = self.args.get('pre_part')
             conf.post_part = self.args.get('post_part')
             conf.part_func = self.args.get('part_func', PART_FUNC_NEW)
+            conf.retention_period = self.args.get('retention_period')
+            conf.ignore_old_events = self.get_arg('ignore_old_events', [0, 1], 0)
         # set row mode and event types to process
         conf.row_mode = self.get_arg('row_mode', ROW_MODES)
         event_types = self.args.get('event_types', '*')
@@ -707,28 +724,22 @@ class Dispatcher(BaseHandler):
                     conf.field_map[tmp[0]] = tmp[0]
                 else:
                     conf.field_map[tmp[0]] = tmp[1]
-        # encoding validator
-        conf.encoding = self.args.get('encoding')
         return conf
 
-    def get_arg(self, name, value_list, default = None):
-        default = default or value_list[0]
-        val = type(default)(self.args.get(name, default))
-        if val not in value_list:
-            raise Exception('Bad argument %s value %r' % (name, val))
-        return val
+    def _validate_hash_key(self):
+        pass # no need for hash key when not sharding
 
     def reset(self):
         """Called before starting to process a batch.
         Should clean any pending data."""
-        BaseHandler.reset(self)
+        ShardHandler.reset(self)
 
     def prepare_batch(self, batch_info, dst_curs):
         """Called on first event for this table in current batch."""
         if self.conf.table_mode != 'ignore':
             self.batch_info = batch_info
             self.dst_curs = dst_curs
-        #BaseHandler.prepare_batch(self, batch_info, dst_curs)
+        ShardHandler.prepare_batch(self, batch_info, dst_curs)
 
     def filter_data(self, data):
         """Process with fields skip and map"""
@@ -753,7 +764,7 @@ class Dispatcher(BaseHandler):
             pkeys = [fmap[p] for p in pkeys if p in fmap]
         return pkeys
 
-    def process_event(self, ev, sql_queue_func, arg):
+    def _process_event(self, ev, sql_queue_func, arg):
         """Process a event.
         Event should be added to sql_queue or executed directly.
         """
@@ -771,31 +782,33 @@ class Dispatcher(BaseHandler):
             raise Exception('Unknown event type: %s' % ev.ev_type)
         # process only operations specified
         if not op in self.conf.event_types:
+            #self.log.debug('dispatch.process_event: ignored event type')
             return
-        self.log.debug('dispatch.process_event: %s/%s' % (
-            ev.ev_type, ev.ev_data))
         if self.pkeys is None:
             self.pkeys = self.filter_pkeys(pkeys.split(','))
         data = self.filter_data(data)
         # prepare split table when needed
         if self.conf.table_mode == 'part':
             dst, part_time = self.split_format(ev, data)
+            if dst in self.ignored_tables:
+                return
             if dst not in self.row_handler.table_map:
                 self.check_part(dst, part_time)
+                if dst in self.ignored_tables:
+                    return
         else:
             dst = self.dest_table
 
         if dst not in self.row_handler.table_map:
             self.row_handler.add_table(dst, LOADERS[self.conf.load_mode],
-                                    self.pkeys, self.conf)
+                                       self.pkeys, self.conf)
         self.row_handler.process(dst, op, data)
-        #BaseHandler.process_event(self, ev, sql_queue_func, arg)
 
     def finish_batch(self, batch_info, dst_curs):
         """Called when batch finishes."""
         if self.conf.table_mode != 'ignore':
             self.row_handler.flush(dst_curs)
-        #BaseHandler.finish_batch(self, batch_info, dst_curs)
+        #ShardHandler.finish_batch(self, batch_info, dst_curs)
 
     def get_part_name(self):
         # if custom part name template given, use it
@@ -874,29 +887,56 @@ class Dispatcher(BaseHandler):
                 have_func = skytools.exists_function(curs, PART_FUNC_OLD, len(PART_FUNC_ARGS))
 
             if have_func:
-                self.log.debug('check_part.exec: func:%s, args: %s' % (pfcall, vals))
+                self.log.debug('check_part.exec: func: %s, args: %s', pfcall, vals)
                 curs.execute(pfcall, vals)
             else:
                 #
-                # Otherwise crete simple clone.
+                # Otherwise create simple clone.
                 #
                 # FixMe: differences from create_partitions():
                 # - check constraints
                 # - inheritance
                 #
-                self.log.debug('part func %s not found, cloning table' % self.conf.part_func)
+                self.log.debug('part func %s not found, cloning table', self.conf.part_func)
                 struct = TableStruct(curs, self.dest_table)
                 struct.create(curs, T_ALL, dst)
 
         exec_with_vals(self.conf.post_part)
-        self.log.info("Created table: %s" % dst)
+        self.log.info("Created table: %s", dst)
+
+        if self.conf.retention_period:
+            dropped = self.drop_obsolete_partitions (self.dest_table, self.conf.retention_period, self.conf.period)
+            if self.conf.ignore_old_events and dropped:
+                for tbl in dropped:
+                    self.ignored_tables.add(tbl)
+                    if tbl in self.row_handler.table_map:
+                        del self.row_handler.table_map[tbl]
+
+    def drop_obsolete_partitions (self, parent_table, retention_period, partition_period):
+        """ Drop obsolete partitions of partition-by-date parent table.
+        """
+        curs = self.dst_curs
+        func = RETENTION_FUNC
+        args = [parent_table, retention_period, partition_period]
+        sql = "select " + func + " (%s, %s, %s)"
+        self.log.debug("func: %s, args: %s", func, args)
+        curs.execute(sql, args)
+        res = [row[0] for row in curs.fetchall()]
+        if res:
+            self.log.info("Dropped tables: %s", ", ".join(res))
+        return res
+
+    def get_copy_condition(self, src_curs, dst_curs):
+        """ Prepare where condition for copy and replay filtering.
+        """
+        return ShardHandler.get_copy_condition(self, src_curs, dst_curs)
 
     def real_copy(self, tablename, src_curs, dst_curs, column_list):
         """do actual table copy and return tuple with number of bytes and rows
-        copyed
+        copied
         """
         _src_cols = _dst_cols = column_list
-        condition = ''
+        condition = self.get_copy_condition (src_curs, dst_curs)
 
         if self.conf.skip_fields:
             _src_cols = [col for col in column_list
@@ -913,7 +953,8 @@ class Dispatcher(BaseHandler):
         else:
             _write_hook = None
 
-        return skytools.full_copy(tablename, src_curs, dst_curs, _src_cols, condition,
+        return skytools.full_copy(tablename, src_curs, dst_curs,
+                                  _src_cols, condition,
                                   dst_tablename = self.dest_table,
                                   dst_column_list = _dst_cols,
                                   write_hook = _write_hook)
