@@ -24,7 +24,7 @@ TABLE_OK           = 5  # setup
 
 SYNC_OK   = 0  # continue with batch
 SYNC_LOOP = 1  # sleep, try again
-SYNC_EXIT = 2  # nothing to do, exit skript
+SYNC_EXIT = 2  # nothing to do, exit script
 
 MAX_PARALLEL_COPY = 8 # default number of allowed max parallel copy processes
 
@@ -54,7 +54,6 @@ class Counter(object):
             elif t.state == TABLE_OK:
                 self.ok += 1
 
-
     def get_copy_count(self):
         return self.copy + self.catching_up + self.wanna_sync + self.do_sync
 
@@ -81,7 +80,7 @@ class TableState(object):
         self.changed = 0
         # position in parallel copy work order
         self.copy_pos = 0
-        # max number of parallel copy processesses allowed
+        # max number of parallel copy processes allowed
         self.max_parallel_copy = MAX_PARALLEL_COPY
 
     def forget(self):
@@ -103,7 +102,7 @@ class TableState(object):
         """Set snapshot."""
         if self.str_snapshot == str_snapshot:
             return
-        self.log.debug("%s: change_snapshot to %s" % (self.name, str_snapshot))
+        self.log.debug("%s: change_snapshot to %s", self.name, str_snapshot)
         self.str_snapshot = str_snapshot
         if str_snapshot:
             self.from_snapshot = skytools.Snapshot(str_snapshot)
@@ -122,8 +121,7 @@ class TableState(object):
         self.state = state
         self.sync_tick_id = tick_id
         self.changed = 1
-        self.log.debug("%s: change_state to %s" % (self.name,
-                                    self.render_state()))
+        self.log.debug("%s: change_state to %s", self.name, self.render_state())
 
     def render_state(self):
         """Make a string to be stored in db."""
@@ -172,8 +170,8 @@ class TableState(object):
     def loaded_state(self, row):
         """Update object with info from db."""
 
-        self.log.debug("loaded_state: %s: %s / %s" % (
-                       self.name, row['merge_state'], row['custom_snapshot']))
+        self.log.debug("loaded_state: %s: %s / %s",
+                       self.name, row['merge_state'], row['custom_snapshot'])
         self.change_snapshot(row['custom_snapshot'], 0)
         self.state = self.parse_state(row['merge_state'])
         self.changed = 0
@@ -203,10 +201,12 @@ class TableState(object):
         return self.max_parallel_copy and\
                     self.copy_pos >= self.max_parallel_copy
 
-    def interesting(self, ev, tick_id, copy_thread):
+    def interesting(self, ev, tick_id, copy_thread, copy_table_name):
         """Check if table wants this event."""
 
         if copy_thread:
+            if self.name != copy_table_name:
+                return False
             if self.state not in (TABLE_CATCHING_UP, TABLE_DO_SYNC):
                 return False
         else:
@@ -235,7 +235,7 @@ class TableState(object):
     def gc_snapshot(self, copy_thread, prev_tick, cur_tick, no_lag):
         """Remove attached snapshot if possible.
 
-        If the event processing is in current moment. the snapshot
+        If the event processing is in current moment, the snapshot
         is not needed beyond next batch.
 
         The logic is needed for mostly unchanging tables,
@@ -278,6 +278,10 @@ class Replicator(CascadedWorker):
         # target database
         db = dbname=somedb host=127.0.0.1
 
+        # extra connect string parameters to add to node public connect strings.
+        # useful values: user= sslmode=
+        #remote_extra_connstr =
+
         # how many tables can be copied in parallel
         #parallel_copies = 1
 
@@ -290,8 +294,24 @@ class Replicator(CascadedWorker):
         # compare: sql to use
         #compare_sql = select count(1) as cnt, sum(hashtext(t.*::text)) as chksum from only _TABLE_ t
         # workaround for hashtext change between 8.3 and 8.4
-        #compare_sql = select count(1) as cnt, sum(('x'||substr(md5(t.*::text),1,16))::bit(64)::bigint) as chksum from only _TABLE_ t     
+        #compare_sql = select count(1) as cnt, sum(('x'||substr(md5(t.*::text),1,16))::bit(64)::bigint) as chksum from only _TABLE_ t
         #compare_fmt = %(cnt)d rows, checksum=%(chksum)s
+
+        ## Parameters for initial node creation: create-root/branch/leaf ##
+
+        # These parameters can be given on either command-line or in config
+        # command-line values override config values.  Those values are
+        # used only during create time, otherwise they are loaded from database.
+
+        # Name for local node.
+        #node_name =
+
+        # public connect string for local node, which other nodes will use
+        # to connect to this one.
+        #public_node_location =
+
+        # connect string for existing node to use as provider
+        #initial_provider_location =
     """
 
     # batch info
@@ -485,14 +505,14 @@ class Replicator(CascadedWorker):
                 else:
                     # regular provider is used
                     if t.name not in pmap:
-                        self.log.warning("Table %s not available on provider" % t.name)
+                        self.log.warning("Table %s not available on provider", t.name)
                         continue
                     pt = pmap[t.name]
                     if pt.state != TABLE_OK: # or pt.custom_snapshot: # FIXME: does snapsnot matter?
-                        self.log.info("Table %s not OK on provider, waiting" % t.name)
+                        self.log.info("Table %s not OK on provider, waiting", t.name)
                         continue
 
-                # dont allow more copies than configured
+                # don't allow more copies than configured
                 if npossible == 0:
                     break
                 npossible -= 1
@@ -504,7 +524,7 @@ class Replicator(CascadedWorker):
                 # failure inbetween
                 self.change_table_state(dst_db, t, TABLE_IN_COPY)
 
-                # the copy _may_ happen immidiately
+                # the copy _may_ happen immediately
                 self.launch_copy(t)
 
                 # there cannot be interesting events in current batch
@@ -513,13 +533,12 @@ class Replicator(CascadedWorker):
 
         return ret
 
-
     def sync_from_copy_thread(self, cnt, src_db, dst_db):
         "Copy thread sync logic."
 
         # somebody may have done remove-table in the meantime
         if self.copy_table_name not in self.table_map:
-            self.log.error("copy_sync: lost table: %s" % self.copy_table_name)
+            self.log.error("copy_sync: lost table: %s", self.copy_table_name)
             return SYNC_EXIT
 
         # This operates on single table
@@ -537,8 +556,8 @@ class Replicator(CascadedWorker):
             elif self.cur_tick < t.sync_tick_id:
                 return SYNC_OK
             else:
-                self.log.error("copy_sync: cur_tick=%d sync_tick=%d" % (
-                                self.cur_tick, t.sync_tick_id))
+                self.log.error("copy_sync: cur_tick=%d sync_tick=%d",
+                                self.cur_tick, t.sync_tick_id)
                 raise Exception('Invalid table state')
         elif t.state == TABLE_WANNA_SYNC:
             # wait for main thread to react
@@ -597,7 +616,7 @@ class Replicator(CascadedWorker):
     def process_remote_event(self, src_curs, dst_curs, ev):
         """handle one event"""
 
-        self.log.debug("New event: id=%s / type=%s / data=%s / extra1=%s" % (ev.id, ev.type, ev.data, ev.extra1))
+        self.log.debug("New event: id=%s / type=%s / data=%s / extra1=%s", ev.id, ev.type, ev.data, ev.extra1)
 
         # set current_event only if processing them one-by-one
         if self.work_state < 0:
@@ -632,9 +651,9 @@ class Replicator(CascadedWorker):
         self.current_event = None
 
     def handle_data_event(self, ev, dst_curs):
-        """handle one truncate event"""
+        """handle one data event"""
         t = self.get_table_by_name(ev.extra1)
-        if not t or not t.interesting(ev, self.cur_tick, self.copy_thread):
+        if not t or not t.interesting(ev, self.cur_tick, self.copy_thread, self.copy_table_name):
             self.stat_increase('ignored_events')
             return
 
@@ -650,12 +669,22 @@ class Replicator(CascadedWorker):
     def handle_truncate_event(self, ev, dst_curs):
         """handle one truncate event"""
         t = self.get_table_by_name(ev.extra1)
-        if not t or not t.interesting(ev, self.cur_tick, self.copy_thread):
+        if not t or not t.interesting(ev, self.cur_tick, self.copy_thread, self.copy_table_name):
             self.stat_increase('ignored_events')
             return
 
         fqname = skytools.quote_fqident(t.dest_table)
 
+        try:
+            p = self.used_plugins[ev.extra1]
+        except KeyError:
+            p = t.get_plugin()
+            self.used_plugins[ev.extra1] = p
+
+        if p.conf.get('ignore_truncate'):
+            self.log.info("ignoring truncate for %s", fqname)
+            return
+
         #
         # Always use CASCADE, because without it the
         # operation cannot work with FKeys, on both
@@ -737,17 +766,6 @@ class Replicator(CascadedWorker):
 
         dst_curs.execute(buf)
 
-    def interesting(self, ev):
-        """See if event is interesting."""
-        if ev.type not in ('I', 'U', 'D', 'R'):
-            raise Exception('bug - bad event type in .interesting')
-        t = self.get_table_by_name(ev.extra1)
-        if not t:
-            return 0
-        if not t.interesting(ev, self.cur_tick, self.copy_thread):
-            return 0
-        return 1
-
     def add_set_table(self, dst_curs, tbl):
         """There was new table added to root, remember it."""
 
@@ -781,7 +799,7 @@ class Replicator(CascadedWorker):
 
         new_list = []
         new_map = {}
-        for row in curs.dictfetchall():
+        for row in curs.fetchall():
             if not row['local']:
                 continue
             t = self.get_table_by_name(row['table_name'])
@@ -824,8 +842,8 @@ class Replicator(CascadedWorker):
             if not t.changed:
                 continue
             merge_state = t.render_state()
-            self.log.info("storing state of %s: copy:%d new_state:%s" % (
-                            t.name, self.copy_thread, merge_state))
+            self.log.info("storing state of %s: copy:%d new_state:%s",
+                            t.name, self.copy_thread, merge_state)
             q = "select londiste.local_set_table_state(%s, %s, %s, %s)"
             curs.execute(q, [self.set_name,
                              t.name, t.str_snapshot, merge_state])
@@ -838,8 +856,8 @@ class Replicator(CascadedWorker):
         self.save_table_state(dst_db.cursor())
         dst_db.commit()
 
-        self.log.info("Table %s status changed to '%s'" % (
-                      tbl.name, tbl.render_state()))
+        self.log.info("Table %s status changed to '%s'",
+                      tbl.name, tbl.render_state())
 
     def get_tables_in_state(self, state):
         "get all tables with specific state"
@@ -857,7 +875,7 @@ class Replicator(CascadedWorker):
         return None
 
     def launch_copy(self, tbl_stat):
-        """Run paraller worker for copy."""
+        """Run parallel worker for copy."""
         self.log.info("Launching copy process")
         script = sys.argv[0]
         conf = self.cf.filename
@@ -867,10 +885,10 @@ class Replicator(CascadedWorker):
         if self.options.quiet:
             cmd.append('-q')
         if self.options.verbose:
-            cmd.append('-v')
+            cmd += ['-v'] * self.options.verbose
 
         # let existing copy finish and clean its pidfile,
-        # otherwise new copy will exit immidiately.
+        # otherwise new copy will exit immediately.
         # FIXME: should not happen on per-table pidfile ???
         copy_pidfile = "%s.copy.%s" % (self.pidfile, tbl_stat.name)
         while skytools.signal_pidfile(copy_pidfile, 0):
@@ -878,11 +896,11 @@ class Replicator(CascadedWorker):
             time.sleep(2)
 
         # launch and wait for daemonization result
-        self.log.debug("Launch args: "+repr(cmd))
+        self.log.debug("Launch args: %r", cmd)
         res = os.spawnvp(os.P_WAIT, script, cmd)
-        self.log.debug("Launch result: "+repr(res))
+        self.log.debug("Launch result: %r", res)
         if res != 0:
-            self.log.error("Failed to launch copy process, result=%d" % res)
+            self.log.error("Failed to launch copy process, result=%d", res)
 
     def sync_database_encodings(self, src_db, dst_db):
         """Make sure client_encoding is same on both side."""
@@ -903,7 +921,7 @@ class Replicator(CascadedWorker):
                 dst_curs.execute("set client_encoding = %s", [src_enc])
 
     def copy_snapshot_cleanup(self, dst_db):
-        """Remove unnecassary snapshot info from tables."""
+        """Remove unnecessary snapshot info from tables."""
         no_lag = not self.work_state
         changes = False
         for t in self.table_list:
@@ -921,7 +939,7 @@ class Replicator(CascadedWorker):
         # restore fkeys -- one at a time
         q = "select * from londiste.get_valid_pending_fkeys(%s)"
         dst_curs.execute(q, [self.set_name])
-        fkey_list = dst_curs.dictfetchall()
+        fkey_list = dst_curs.fetchall()
         for row in fkey_list:
             self.log.info('Creating fkey: %(fkey_name)s (%(from_table)s --> %(to_table)s)' % row)
             q2 = "select londiste.restore_table_fkey(%(from_table)s, %(fkey_name)s)"
@@ -937,7 +955,7 @@ class Replicator(CascadedWorker):
         dst_curs = dst_db.cursor()
         q = "select * from londiste.find_table_fkeys(%s)"
         dst_curs.execute(q, [table_name])
-        fkey_list = dst_curs.dictfetchall()
+        fkey_list = dst_curs.fetchall()
         for row in fkey_list:
             self.log.info('Dropping fkey: %s' % row['fkey_name'])
             q2 = "select londiste.drop_table_fkey(%(from_table)s, %(fkey_name)s)"
@@ -979,4 +997,3 @@ class Replicator(CascadedWorker):
 if __name__ == '__main__':
     script = Replicator(sys.argv[1:])
     script.start()
-
