@@ -3,9 +3,18 @@
 
 """
 
-import sys, os, signal, optparse, time, errno, select
-import logging, logging.handlers, logging.config
+import errno
+import logging
+import logging.config
+import logging.handlers
+import optparse
+import os
+import select
+import signal
+import sys
+import time
 
+import psycopg2
 import skytools
 import skytools.skylog
 
@@ -126,7 +135,8 @@ def _init_log(job_name, service_name, cf, log_level, is_daemon):
         skytools.skylog.set_service_name(service_name, job_name)
 
         # load general config
-        flist = ['skylog.ini', '~/.skylog.ini', '/etc/skylog.ini']
+        flist = cf.getlist('skylog_locations',
+                           ['skylog.ini', '~/.skylog.ini', '/etc/skylog.ini'])
         for fn in flist:
             fn = os.path.expanduser(fn)
             if os.path.isfile(fn):
@@ -189,7 +199,7 @@ class BaseScript(object):
 
     Config template::
 
-        ## Parameters for skytools.DBScript ##
+        ## Parameters for skytools.BaseScript ##
 
         # how many seconds to sleep between work loops
         # if missing or 0, then instead sleeping, the script will exit
@@ -211,6 +221,9 @@ class BaseScript(object):
         #   2 - always enabled
         #use_skylog = 0
 
+        # where to find skylog.ini
+        #skylog_locations = skylog.ini, ~/.skylog.ini, /etc/skylog.ini
+
         # how many seconds to sleep after catching a exception
         #exception_sleep = 20
     """
@@ -223,7 +236,7 @@ class BaseScript(object):
     # >0 - sleep time if work() requests sleep
     # 0  - exit if work requests sleep
     # <0 - run work() once [same as looping=0]
-    loop_delay = 0
+    loop_delay = 1.0
 
     # 0 - run work() once
     # 1 - run work() repeatedly
@@ -377,11 +390,11 @@ class BaseScript(object):
         parse command line arguments.
 
         Note that it can be overridden both directions - either DBScript
-        will initialize a instance and passes to user code or user can
+        will initialize an instance and pass it to user code or user can
         initialize and then pass to DBScript.init_optparse().
 
         @param parser: optional OptionParser() instance,
-               where DBScript should attachs its own arguments.
+               where DBScript should attach its own arguments.
         @return: initialized OptionParser() instance.
         """
         if parser:
@@ -389,6 +402,7 @@ class BaseScript(object):
         else:
             p = optparse.OptionParser()
             p.set_usage("%prog [options] INI")
+
         # generic options
         p.add_option("-q", "--quiet", action="store_true",
                      help = "log only errors and warnings")
@@ -461,8 +475,11 @@ class BaseScript(object):
             self.log.info ("Config reloaded")
         self.job_name = self.cf.get("job_name")
         self.pidfile = self.cf.getfile("pidfile", '')
-        self.loop_delay = self.cf.getfloat("loop_delay", 1.0)
+        self.loop_delay = self.cf.getfloat("loop_delay", self.loop_delay)
         self.exception_sleep = self.cf.getfloat("exception_sleep", 20)
+        self.exception_quiet = self.cf.getlist("exception_quiet", [])
+        self.exception_grace = self.cf.getfloat("exception_grace", 5*60)
+        self.exception_reset = self.cf.getfloat("exception_reset", 15*60)
 
     def hook_sighup(self, sig, frame):
         "Internal SIGHUP handler.  Minimal code here."
@@ -555,12 +572,14 @@ class BaseScript(object):
 
         return state
 
+    last_func_fail = None
     def run_func_safely(self, func, prefer_looping = False):
         "Run users work function, safely."
-        cname = None
-        emsg = None
         try:
-            return func()
+            r = func()
+            if self.last_func_fail and time.time() > self.last_func_fail + self.exception_reset:
+                self.last_func_fail = None
+            return r
         except UsageError, d:
             self.log.error(str(d))
             sys.exit(1)
@@ -583,7 +602,12 @@ class BaseScript(object):
             self.reset()
             sys.exit(1)
         except Exception, d:
-            self.send_stats()
+            try: # this may fail too
+                self.send_stats()
+            except:
+                pass
+            if self.last_func_fail is None:
+                self.last_func_fail = time.time()
             emsg = str(d).rstrip()
             self.reset()
             self.exception_hook(d, emsg)
@@ -602,6 +626,10 @@ class BaseScript(object):
             if ex.errno != errno.EINTR:
                 raise
 
+    def _is_quiet_exception(self, ex):
+        return ((self.exception_quiet == ["ALL"] or ex.__class__.__name__ in self.exception_quiet)
+                and self.last_func_fail and time.time() < self.last_func_fail + self.exception_grace)
+
     def exception_hook(self, det, emsg):
         """Called on after exception processing.
 
@@ -610,8 +638,11 @@ class BaseScript(object):
         @param det: exception details
         @param emsg: exception msg
         """
-        self.log.exception("Job %s crashed: %s" % (
-                   self.job_name, emsg))
+        lm = "Job %s crashed: %s" % (self.job_name, emsg)
+        if self._is_quiet_exception(det):
+            self.log.warning(lm)
+        else:
+            self.log.exception(lm)
 
     def work(self):
         """Here should user's processing happen.
@@ -691,8 +722,17 @@ class DBScript(BaseScript):
     def set_database_defaults(self, dbname, **kwargs):
         self._db_defaults[dbname] = kwargs
 
+    def add_connect_string_profile(self, connstr, profile):
+        """Add extra profile info to connect string.
+        """
+        if profile:
+            extra = self.cf.get("%s_extra_connstr" % profile, '')
+            if extra:
+                connstr += ' ' + extra
+        return connstr
+
     def get_database(self, dbname, autocommit = 0, isolation_level = -1,
-                     cache = None, connstr = None):
+                     cache = None, connstr = None, profile = None):
         """Load cached database connection.
 
         User must not store it permanently somewhere,
@@ -720,14 +760,16 @@ class DBScript(BaseScript):
             params['max_age'] = max_age
 
         if cache in self.db_cache:
+            dbc = self.db_cache[cache]
             if connstr is None:
                 connstr = self.cf.get(dbname, '')
-            dbc = self.db_cache[cache]
             if connstr:
+                connstr = self.add_connect_string_profile(connstr, profile)
                 dbc.check_connstr(connstr)
         else:
             if not connstr:
                 connstr = self.cf.get(dbname)
+            connstr = self.add_connect_string_profile(connstr, profile)
 
             # connstr might contain password, it is not a good idea to log it
             filtered_connstr = connstr
@@ -782,8 +824,12 @@ class DBScript(BaseScript):
             sql = getattr(curs, 'query', None) or '?'
             if len(sql) > 200: # avoid logging londiste huge batched queries
                 sql = sql[:60] + " ..."
-            self.log.exception("Job %s got error on connection '%s': %s.   Query: %s" % (
-                self.job_name, cname, emsg, sql))
+            lm = "Job %s got error on connection '%s': %s.   Query: %s" % (
+                self.job_name, cname, emsg, sql)
+            if self._is_quiet_exception(d):
+                self.log.warning(lm)
+            else:
+                self.log.exception(lm)
         else:
             BaseScript.exception_hook(self, d, emsg)
 
@@ -902,6 +948,48 @@ class DBScript(BaseScript):
             # error is already logged
             sys.exit(1)
 
+    def execute_with_retry (self, dbname, stmt, args, exceptions = None):
+        """ Execute SQL and retry if it fails.
+        Return number of retries and current valid cursor, or raise an exception.
+        """
+        sql_retry = self.cf.getbool("sql_retry", False)
+        sql_retry_max_count = self.cf.getint("sql_retry_max_count", 10)
+        sql_retry_max_time = self.cf.getint("sql_retry_max_time", 300)
+        sql_retry_formula_a = self.cf.getint("sql_retry_formula_a", 1)
+        sql_retry_formula_b = self.cf.getint("sql_retry_formula_b", 5)
+        sql_retry_formula_cap = self.cf.getint("sql_retry_formula_cap", 60)
+        elist = exceptions or (psycopg2.OperationalError,)
+        stime = time.time()
+        tried = 0
+        dbc = None
+        while True:
+            try:
+                if dbc is None:
+                    if dbname not in self.db_cache:
+                        self.get_database(dbname, autocommit=1)
+                    dbc = self.db_cache[dbname]
+                    if dbc.isolation_level != skytools.I_AUTOCOMMIT:
+                        raise skytools.UsageError ("execute_with_retry: autocommit required")
+                else:
+                    dbc.reset()
+                curs = dbc.get_connection(dbc.isolation_level).cursor()
+                curs.execute (stmt, args)
+                break
+            except elist, e:
+                if not sql_retry or tried >= sql_retry_max_count or time.time() - stime >= sql_retry_max_time:
+                    raise
+                self.log.info("Job %s got error on connection %s: %s" % (self.job_name, dbname, e))
+            except:
+                raise
+            # y = a + bx , apply cap
+            y = sql_retry_formula_a + sql_retry_formula_b * tried
+            if sql_retry_formula_cap is not None and y > sql_retry_formula_cap:
+                y = sql_retry_formula_cap
+            tried += 1
+            self.log.info("Retry #%i in %i seconds ...", tried, y)
+            self.sleep(y)
+        return tried, curs
+
     def listen(self, dbname, channel):
         """Make connection listen for specific event channel.
 
@@ -941,7 +1029,6 @@ class DBCachedConn(object):
         self.conn = None
         self.conn_time = 0
         self.max_age = max_age
-        self.autocommit = -1
         self.isolation_level = -1
         self.verbose = verbose
         self.setup_func = setup_func
