@@ -4,11 +4,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: londiste.version(0)
 --
---      Returns version string for londiste.  ATM it is based on SkyTools version
---      and only bumped when database code changes.
+--      Returns version string for londiste.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '3.1.3';
+    return '3.2';
 end;
 $$ language plpgsql;
 
