@@ -3,11 +3,11 @@ returns text as $$
 -- ----------------------------------------------------------------------
 -- Function: pgq.version(0)
 --
---      Returns verison string for pgq.  ATM its SkyTools version
---      that is only bumped when PGQ database code changes.
+--      Returns version string for pgq.  ATM it is based on SkyTools
+--      version and only bumped when database code changes.
 -- ----------------------------------------------------------------------
 begin
-    return '2.1.8';
+    return '3.2';
 end;
 $$ language plpgsql;
 
