@@ -3,6 +3,7 @@
 import sys, time
 import skytools.psycopgwrapper
 import skytools._cquoting, skytools._pyquoting
+from decimal import Decimal
 
 # create a DictCursor row
 class fake_cursor:
@@ -17,12 +18,12 @@ def regtest(name, func, cases):
     for dat, res in cases:
         res2 = func(dat)
         if res != res2:
-            print "failure: %s(%s) = %s (expected %s)" % (name, repr(dat), repr(res2), repr(res))
+            print("failure: %s(%s) = %s (expected %s)" % (name, repr(dat), repr(res2), repr(res)))
             bad += 1
     if bad:
-        print "%-20s: failed" % name
+        print("%-20s: failed" % name)
     else:
-        print "%-20s: OK" % name
+        print("%-20s: OK" % name)
             
 
 sql_literal = [
@@ -32,6 +33,7 @@ sql_literal = [
     [r"a\'b", r"E'a\\''b'"],
     [1, "'1'"],
     [True, "'True'"],
+    [Decimal(1), "'1'"],
 ]
 regtest("quote_literal/c", skytools._cquoting.quote_literal, sql_literal)
 regtest("quote_literal/py", skytools._pyquoting.quote_literal, sql_literal)
@@ -44,6 +46,7 @@ sql_copy = [
     [1, "1"],
     [True, "True"],
     [u"qwe", "qwe"],
+    [Decimal(1), "1"],
 ]
 regtest("quote_copy/c", skytools._cquoting.quote_copy, sql_copy)
 regtest("quote_copy/py", skytools._pyquoting.quote_copy, sql_copy)
@@ -75,6 +78,7 @@ t_urlenc = [
     [{'qwe': 1, u'zz': u"qwe"}, "qwe=1&zz=qwe"],
     [{'a': '\000%&'}, "a=%00%25%26"],
     [dbrow, 'data=value&id=123'],
+    [{'a': Decimal("1")}, "a=1"],
 ]
 regtest("db_urlencode/c", skytools._cquoting.db_urlencode, t_urlenc)
 regtest("db_urlencode/py", skytools._pyquoting.db_urlencode, t_urlenc)
